;;;
;;; This program implements the Wang algorithm for proving
;;; formulas of the propositional calculus.
;;;
;;;------------------------
;;;
;;; The driver routine.
;;;
prove-or-refute[conjecture] =
  {hlist = hypotheses[conjecture];
   glist = goals[conjecture] :
   [have-common-props[hlist; glist] -->
      make-simple-proof[conjecture; first-common-prop[hlist; glist]];
    has-composite-formula[hlist] -->
      apply-hypothesis-reduction
        [conjecture; first-composite-formula[hlist];
         remove-first-composite-formula[hlist]; glist];
    has-composite-formula[glist] -->
      apply-goal-reduction
        [conjecture; first-composite-formula[glist]; hlist;
         remove-first-composite-formula[glist]];
    otherwise --> make-simple-refutation[conjecture]]}
;;;
;;; Check for a common proposition in the hypothesis and goal lists.
;;;
have-common-props[hlist; glist] =
  [endp[hlist] --> F;
   proposition-p[first[hlist]] -->
     [hasprop[first[hlist]; glist] --> T;
      otherwise --> have-common-props[rest[hlist]; glist]];
   otherwise --> have-common-props[rest[hlist]; glist]]

hasprop[p; list] =
  [endp[list] --> F;
   proposition-p[first[list]] -->
     [eq[prop-name[p]; prop-name[first[list]]] --> T;
      otherwise --> hasprop[p; rest[list]]];
   otherwise --> hasprop[p; rest[list]]]

first-common-prop[hlist; glist] =
  [proposition-p[first[hlist]] -->
     [hasprop[first[hlist]; glist] --> first[hlist];
      otherwise --> first-common-prop[rest[hlist]; glist]];
   otherwise --> first-common-prop[rest[hlist]; glist]]
;;;
;;; Are there composite formulas in a list of formulas?
;;; If so, extract the first one.
;;;
has-composite-formula[list] =
  [endp[list] --> F;
   proposition-p[first[list]] --> has-composite-formula[rest[list]];
   otherwise --> T]

first-composite-formula[list] =
  [proposition-p[first[list]] --> first-composite-formula[rest[list]];
   otherwise --> first[list]]

remove-first-composite-formula[list] =
  [proposition-p[first[list]] -->
     cons[first[list]; remove-first-composite-formula[rest[list]]];
   otherwise --> rest[list]]
;;;
;;; The following routines apply a reduction of hypothesis
;;; and goal formulas respectively.
;;;
cons2[e1; e2; list] = cons[e1; cons[e2; list]]

apply-hypothesis-reduction[conjecture; h; newhlist; glist] =
  [negation-p[h] -->
     single-reduction
       [conjecture; h; rule2A;
        make-conjecture[newhlist; cons[negend[h]; glist]]];
   conjunction-p[h] -->
     single-reduction
       [conjecture; h; rule3A;
        make-conjecture
          [cons2[conjunct1[h]; conjunct2[h]; newhlist]; glist]];
   disjunction-p[h] -->
     double-reduction
       [conjecture; h; rule4A;
        make-conjecture[cons[disjunct1[h]; newhlist]; glist];
        make-conjecture[cons[disjunct2[h]; newhlist]; glist]];
   implication-p[h] -->
     double-reduction
       [conjecture; h; rule5A;
        make-conjecture[newhlist; cons[antecedent[h]; glist]];
        make-conjecture[cons[consequent[h]; newhlist]; glist]];
   equivalence-p[h] -->
     double-reduction
       [conjecture; h; rule6A;
        make-conjecture
          [cons2[condition1[h]; condition2[h]; newhlist]; glist];
        make-conjecture
          [newhlist; cons2[condition1[h]; condition2[h]; glist]]]]

apply-goal-reduction[conjecture; g; hlist; newglist] =
  [negation-p[g] -->
     single-reduction
       [conjecture; g; rule2B;
        make-conjecture[cons[negend[g]; hlist]; newglist]];
   conjunction-p[g] -->
     double-reduction
       [conjecture; g; rule3B;
        make-conjecture[hlist; cons[conjunct1[g]; newglist]];
        make-conjecture[hlist; cons[conjunct2[g]; newglist]]];
   disjunction-p[g] -->
     single-reduction
       [conjecture; g; rule4B;
        make-conjecture
          [hlist; cons2[disjunct1[g]; disjunct2[g]; newglist]]];
   implication-p[g] -->
     single-reduction
       [conjecture; g; rule5B;
        make-conjecture
          [cons[antecedent[g]; hlist];
           cons[consequent[g]; newglist]]];
   equivalence-p[g] -->
     double-reduction
       [conjecture; g; rule6B;
        make-conjecture
          [cons[condition1[g]; hlist]; cons[condition2[g]; newglist]];
        make-conjecture
          [cons[condition2[g]; hlist];
           cons[condition1[g]; newglist]]]]
;;;
;;; Perform a proof step involving a single reduction.
;;;
single-reduction[conjecture; formula; rule; newconjecture] =
  {subproof-or-refutation = prove-or-refute[newconjecture] :
   [proof-p[subproof-or-refutation] -->
      make-linear-proof
        [conjecture; rule; formula; subproof-or-refutation];
    otherwise -->
      make-linear-refutation
        [conjecture; rule; formula; subproof-or-refutation]]}
;;;
;;; Perform a proof step involving a double reduction.
;;;
double-reduction
  [conjecture; formula; rule; newconjecture1; newconjecture2] =
  {subproof-or-refutation1 = prove-or-refute[newconjecture1] :
   [proof-p[subproof-or-refutation1] -->
      {subproof-or-refutation2 = prove-or-refute[newconjecture2] :
       [proof-p[subproof-or-refutation2] -->
          make-double-proof
            [conjecture; rule; formula; subproof-or-refutation1;
             subproof-or-refutation2];
        otherwise -->
          make-linear-refutation
            [conjecture; rule; formula; subproof-or-refutation2]]};
    otherwise -->
      make-linear-refutation
        [conjecture; rule; formula; subproof-or-refutation1]]}
